from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite


class Catalog(object):

    @property
    def catalog(self):
        return getToolByName(getSite(), 'portal_catalog')

    def get_indexed_value_for(self, obj, index_name):
        rid = self.catalog.getrid('/'.join(obj.getPhysicalPath()))
        indexed_data = self.catalog._catalog.getIndexDataForRID(rid)
        return indexed_data.get(index_name)

    def get_allowed_roles_and_users(self, obj):
        return self.get_indexed_value_for(obj, 'allowedRolesAndUsers')



from collective.indexing.queue import getQueue
from ftw.builder import Builder
from ftw.builder import create
from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from ftw.solr.tests.pages import Catalog
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase
import transaction


class TestReindexingSecurity(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

    def get_indexing_queue_length(self):
        return getQueue().length()

    def folder_builder(self):
        return Builder('folder')

    def test_indexes_are_updated_recursively(self):
        folder = create(self.folder_builder())
        subfolder = create(self.folder_builder().within(folder))

        self.assertEquals(
            {'folder': ['Anonymous'],
             'subfolder': ['Anonymous']},

            {'folder': Catalog().get_allowed_roles_and_users(folder),
             'subfolder': Catalog().get_allowed_roles_and_users(subfolder)})

        folder.manage_permission('View', roles=['Manager'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEquals(2, self.get_indexing_queue_length())
        transaction.commit()

        self.assertEquals(
            {'folder': set(['Manager']),
             'subfolder': set(['Manager'])},

            {'folder': set(Catalog().get_allowed_roles_and_users(folder)),
             'subfolder': set(Catalog().get_allowed_roles_and_users(subfolder))})

    def test_resursive_indexing_is_not_done_when_subobjects_do_not_acquire(self):
        folder = create(self.folder_builder())
        folder.manage_permission('View', roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        subfolder = create(self.folder_builder().within(folder))
        subfolder.manage_permission('View', roles=['Reader'], acquire=False)
        subfolder.reindexObjectSecurity()
        transaction.commit()

        self.assertEquals(
            {'folder': ['Reader'],
             'subfolder': ['Reader']},

            {'folder': Catalog().get_allowed_roles_and_users(folder),
             'subfolder': Catalog().get_allowed_roles_and_users(subfolder)})

        self.assertEquals(0, self.get_indexing_queue_length(),
                          'Queue should be empty at this point.')
        folder.manage_permission('View', roles=['Manager'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEquals(1, self.get_indexing_queue_length(),
                          'Only the folder should be reindexed at this point,'
                          ' since the subfolder does not acquire the View permission.')
        transaction.commit()

        self.assertEquals(
            {'folder': ['Manager'],
             'subfolder': ['Reader']},

            {'folder': Catalog().get_allowed_roles_and_users(folder),
             'subfolder': Catalog().get_allowed_roles_and_users(subfolder)})


class TestDexterityReindexingSecurity(TestReindexingSecurity):

    def folder_builder(self):
        return Builder('dexterity folder')



from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest2 import TestCase


class TestReindexObjectSecurityPatches(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_CatalogMultiplex_reindexObjectSecurity_is_patched(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])

        # Test base class CatalogMultiplex is patched
        from Products.Archetypes.CatalogMultiplex import CatalogMultiplex
        self.assertEquals(CatalogMultiplex.reindexObjectSecurity.__func__.func_name,
                          'ftw_solr_CatalogMultiplex_reindexObjectSecurity')

        # Test actual objects inheriting from CatalogMultiplex are patched
        doc = portal[portal.invokeFactory('Document', 'doc1', title=u"Doc1")]
        self.assertEquals(doc.reindexObjectSecurity.__func__.func_name,
                          'ftw_solr_CatalogMultiplex_reindexObjectSecurity')

    def test_CatalogAware_reindexObjectSecurity_is_patched(self):
        from Products.CMFCore.CMFCatalogAware import CatalogAware
        self.assertEquals(CatalogAware.reindexObjectSecurity.__func__.func_name,
                          'ftw_solr_CatalogAware_reindexObjectSecurity')



from collective.solr.mangler import cleanupQueryParameters
from collective.solr.mangler import optimizeQueryParameters
from collective.solr.mangler import subtractQueryParameters
from collective.solr.queryparser import quote
from collective.solr.search import languageFilter
from collective.solr.utils import isWildCard
from collective.solr.utils import prepare_wildcard
from collective.solr.utils import prepareData
from ftw.solr.patches.mangler import mangleQuery
from logging import getLogger
from Missing import MV


logger = getLogger('collective.solr.search')


def buildQueryAndParameters(self, default=None, **args):
    """ helper to build a querystring for simple use-cases """
    schema = self.getManager().getSchema() or {}
    config = self.getConfig()

    params = subtractQueryParameters(args)
    params = cleanupQueryParameters(params, schema)

    languageFilter(args)
    prepareData(args)
    mangleQuery(args, config, schema)

    logger.debug('building query for "%r", %r', default, args)
    defaultSearchField = getattr(schema, 'defaultSearchField', None)

    if default is not None and defaultSearchField is not None:
        args[None] = default

    query = {}

    for name, value in sorted(args.items()):
        field = schema.get(name or defaultSearchField, None)
        if field is None or not field.indexed:
            logger.info(
                'dropping unknown search attribute "%s" '
                ' (%r) for query: %r', name, value, args
            )
            continue
        if isinstance(value, bool):
            value = str(value).lower()
        elif not value:     # solr doesn't like empty fields (+foo:"")
            if not name:
                continue
            logger.info(
                'empty search term form "%s:%s", aborting buildQuery' % (
                    name,
                    value
                )
            )
            return {}, params
        elif field.class_ == 'solr.BoolField':
            if not isinstance(value, (tuple, list)):
                value = [value]
            falses = '0', 'False', MV
            true = lambda v: bool(v) and v not in falses
            value = set(map(true, value))
            if not len(value) == 1:
                assert len(value) == 2      # just to make sure
                continue                    # skip when "true or false"
            value = str(value.pop()).lower()
        elif isinstance(value, (tuple, list)):
            # list items should be treated as literals, but
            # nevertheless only get quoted when necessary
            def quoteitem(term):
                if isinstance(term, unicode):
                    term = term.encode('utf-8')
                quoted = quote(term)
                if not quoted.startswith('"') and not quoted == term:
                    quoted = quote('"' + term + '"')
                return quoted
            value = '(%s)' % ' OR '.join(map(quoteitem, value))
        elif isinstance(value, set):        # sets are taken literally
            if len(value) == 1:
                query[name] = ''.join(value)
            else:
                query[name] = '(%s)' % ' OR '.join(value)
            if '/' in query[name]:
                query[name] = query[name].replace('/', '\\/')
            continue
        elif isinstance(value, basestring):
            if field.class_ == 'solr.TextField':
                if isWildCard(value):
                    value = prepare_wildcard(value)
                value = quote(value, textfield=True)
                # if we have an intra-word hyphen, we need quotes
                if '\\-' in value or '\\+' in value:
                    if value[0] != '"':
                        value = '"%s"' % value
            else:
                value = quote(value)
            if not value:   # don't search for empty strings, even quoted
                continue
        else:
            logger.info(
                'skipping unsupported value "%r" (%s)', value, name
            )
            continue
        if name is None:
            if value and value[0] not in '+-':
                value = '+%s' % value
        else:
            value = '+%s:%s' % (name, value)

        query[name] = value
    logger.debug('built query "%s"', query)

    if query:
        optimizeQueryParameters(query, params)
    return query, params



from collective.solr.browser.facets import facetParameters
from collective.solr.interfaces import ISolrConnectionManager
from itertools import islice
from math import ceil
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import queryUtility
from zope.publisher.browser import BrowserView
import json
import requests


CLOUD_TERMS_FIELD = 'wordCloudTerms'


class WordCloudView(BrowserView):
    """BrowserView to fetch data for and render a Word Cloud.

    Currently term frequency across all documents indexed in Solr is used as
    basis for the terms weights.
    """
    fragment_template = ViewPageTemplateFile('wordcloud_fragment.pt')
    debug_template = ViewPageTemplateFile('wordcloud_debug.pt')

    def get_scale(self, weight):
        """Scale term size according to its weight.
        """
        n = self.options['num_sizes']
        # Scale linearly for now
        scale = ceil(weight * n) / n
        return scale

    def get_style(self, weight):
        """Given a specific term weight, return CSS styles that format an <a>
        element accordingly.
        """
        scale = self.get_scale(weight)

        # Size in percent of inherited font size
        size = int(100 * scale)

        # Scale by configurable constant factor
        size *= self.options['scale_factor']

        return 'font-size: %s%%; border-bottom:0; text-decoration: none' % size

    def get_solr_terms_handler_url(self):
        """Construct the URL to the Solr 'terms' request handler.
        """
        manager = queryUtility(ISolrConnectionManager)
        conn = manager.getConnection()
        if conn is None:
            raise Exception("Couldn't get a Solr connection. Please make sure "
                            "Solr is installed and activated.")
        url = "http://%s%s/terms" % (conn.host, conn.solrBase)
        return url

    def get_terms(self):
        """Return a dictionary of term:count pairs, where `term` is the term's
        name and `count` the number of times the term appears across all
        documents.
        """
        num_words = self.options['num_words']
        params = {'terms.fl': CLOUD_TERMS_FIELD,
                  'terms.limit': num_words,
                  'wt': 'json'}
        url = self.get_solr_terms_handler_url()
        response = requests.post(url, data=params)
        data = response.content
        results = json.loads(data)

        response_header = results['responseHeader']
        if not response_header['status'] == 0:
            # TODO Better exception handling
            raise Exception(str(response_header))

        terms = results['terms'][CLOUD_TERMS_FIELD]

        # terms is a list of [key1, val1, key2, val2, ...]
        # so we extract keys and values and zip() them into a dict
        keys = islice(terms, 0, None, 2)
        values = islice(terms, 1, None, 2)
        cloud_terms = {}
        for pair in zip(keys, values):
            term, count = pair
            cloud_terms[term] = count

        return cloud_terms

    def get_weighted_terms(self):
        """Return a list of term dictionaries including the relative weight
        of the term.
        """
        terms = self.get_terms()
        max_count = max(terms.values())

        weighted_terms = []
        for term, count in terms.items():
            weight = float(count) / max_count
            style = self.get_style(weight)
            weighted_terms.append({'name': term,
                                   'count': count,
                                   'weight': weight,
                                   'style': style})
        # Sort alphabetically by term name by default.
        # If further post-processing in python is required, this will need to
        # be changed to sort by count (reversed).
        weighted_terms.sort(key=lambda x: x['name'])
        return weighted_terms

    def search_url(self):
        """Construct the URL that will be used to link cloud terms to a search
        for that term. Assumes the search term will be appended at the end.
        """
        portal_url = getToolByName(self.context, 'portal_url')
        base_url = portal_url.getPortalObject().absolute_url()
        url = "%s/@@search?" % base_url

        # Add faceting parameters
        facets, dependencies = facetParameters(self.context, self.request)
        url += '&facet=true'
        for facet in facets:
            url += '&facet.field=%s' % facet

        url += '&SearchableText='
        return url

    def _parse_option(self, key, kwargs, default):
        """Parses options needed for the WordCloudView and stores them in the
        self.options dict.

        Values received in request.form have the highest priority and override
        options passed in as keyword arguments to the call() method. If no
        value for an option is supplied in the request, the kwarg is used. If
        there's no kwarg either, we fall back to `default`.

        After parsing the options and populating self.options, kwargs should be
        updated with self.options.
        """
        request_value = self.request.form.get(key)
        if request_value not in (None, ''):
            value = request_value
        else:
            value = kwargs.get(key, default)
        self.options[key] = value

    def __call__(self, *args, **kwargs):
        self.options = {}
        self._parse_option('num_words', kwargs, 10)
        self._parse_option('num_sizes', kwargs, 5)
        self._parse_option('scale_factor', kwargs, 2.0)
        kwargs.update(self.options)

        if self.request.form.get('raw'):
            # Raw output - return weighted terms as pretty printed JSON
            return json.dumps(self.get_weighted_terms(),
                              sort_keys=True,
                              indent=4,
                              separators=(',', ': '))

        elif self.request.form.get('debug'):
            # Debug output - wrap the fragment with a debug template to get a
            # complete HTML document
            return self.debug_template(self, **kwargs)

        # Standard fragment template - only for inclusion in an existing HTML
        # document (as a portlet for example.)
        return self.fragment_template(self, **kwargs)



from ftw.solr.browser.livesearch import FtwSolrLiveSearchReplyView
from ftw.solr.testing import SOLR_INTEGRATION_TESTING
from unittest2 import TestCase


class TestGetShowMoreLink(TestCase):

    layer = SOLR_INTEGRATION_TESTING

    def setUp(self):
        request = self.layer['request']
        self.livesearch = FtwSolrLiveSearchReplyView(object, request)
        self.livesearch.facet_params = 'facet.field=portal_type'
        self.livesearch.searchterms = 'james'

    def test_has_linktext(self):
        result = self.livesearch.get_show_more_item()

        self.assertEqual('Show all items', result.get('title'))

    def test_has_search_term(self):
        result = self.livesearch.get_show_more_item()

        self.assertIn('SearchableText=james', result.get('url'))

    def test_has_path_parameter_if_path_is_set(self):
        self.livesearch.request.form.update({'path': '/path/to/context', })
        result = self.livesearch.get_show_more_item()

        self.assertIn('path=%2Fpath%2Fto%2Fcontext', result.get('url'))

    def test_has_no_path_parameter_if_no_path_is_set(self):
        result = self.livesearch.get_show_more_item()

        self.assertNotIn('path=', result.get('url'))

    def test_has_facet_params(self):
        result = self.livesearch.get_show_more_item()

        self.assertIn('facet.field=portal_type', result.get('url'))



# -*- coding: utf-8 -*-
from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from unittest2 import TestCase
from zope.component import getMultiAdapter, queryUtility
from zope.event import notify
from zope.traversing.interfaces import BeforeTraverseEvent
from collective.solr.interfaces import ISolrConnectionManager
from collective.solr.tests.utils import fakehttp
from ftw.solr.tests.utils import getData
import json


class TestSuggestView(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_suggest(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        # Setup solr suggest response
        response = getData('suggest_response.txt')
        proc = queryUtility(ISolrConnectionManager)
        proc.setHost(active=True)
        conn = proc.getConnection()
        fakehttp(conn, response)

        request.form.update({'term': 'ab'})
        view = getMultiAdapter((portal, request), name=u'suggest-terms')
        suggestions = json.loads(view())

        self.assertEquals(len(suggestions), 9)
        self.assertEquals(suggestions[0]['value'], 'aber')
        self.assertEquals(suggestions[0]['label'], 'aber')

    def test_no_suggest(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        # Setup solr suggest response
        response = getData('nosuggest_response.txt')
        proc = queryUtility(ISolrConnectionManager)
        proc.setHost(active=True)
        conn = proc.getConnection()
        fakehttp(conn, response)

        request.form.update({'term': 'abx'})
        view = getMultiAdapter((portal, request), name=u'suggest-terms')
        suggestions = json.loads(view())
        self.assertEquals(len(suggestions), 0)



from ftw.upgrade import UpgradeStep


class RemoveCollectiveSolrStylingFromFtwTheming(UpgradeStep):
    """Remove collective solr styling from ftw.theming.
    """

    def __call__(self):
        self.install_upgrade_profile()



from collective.solr import flare
from collective.solr.interfaces import IFlare
from collective.solr.interfaces import ISolrFlare
from ftw.solr.interfaces import IFtwSolrLayer
from zope.component import adapts
from zope.interface import implements


# Customized PloneFlare which makes sure string values are returned as
# byte strings.
class PloneFlare(flare.PloneFlare):
    """A solr brain for search results"""
    implements(IFlare)
    adapts(ISolrFlare, IFtwSolrLayer)

    def __getattr__(self, name):
        """Look up attributes in dict"""
        value = super(PloneFlare, self).__getattr__(name)
        if isinstance(value, unicode):
            return value.encode('utf8')
        else:
            return value

    def getURL(self):
        if self.getRemoteUrl and self.getRemoteUrl.startswith('http'):
            return self.getRemoteUrl
        return super(PloneFlare, self).getURL()




from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.parser import SolrResponse
from ftw.solr.browser.facets import SearchFacetsView
from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from ftw.solr.tests.utils import getData
from unittest2 import TestCase
from zope.component import queryUtility


class TestFacetsView(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_facets_order(self):
        portal = self.layer['portal']
        request = self.layer['request']
        request.form.update({'facet_field': ['type', 'section', 'topics']})
        response = SolrResponse(getData('facets_response.xml'))
        view = SearchFacetsView(portal, request)
        view.kw = dict(results=response)

        config = queryUtility(ISolrConnectionConfig)
        config.facets = ['type', 'section', 'topics']
        facets = view.facets()
        self.assertEquals(['type', 'section', 'topics'],
            [facets[0]['title'], facets[1]['title'], facets[2]['title']],
            msg='Wrong facet order.')

        config.facets = ['section', 'topics', 'type']
        facets = view.facets()
        self.assertEquals(['section', 'topics', 'type'],
            [facets[0]['title'], facets[1]['title'], facets[2]['title']],
            msg='Wrong facet order.')



from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.manager import SolrConnectionConfig
from collective.solr.parser import SolrSchema, SolrField
from collective.solr.tests.utils import getData
from ftw.solr.patches.mangler import cleanupQueryParameters
from ftw.solr.patches.mangler import subtractQueryParameters
from ftw.solr.patches.mangler import leading_wildcards
from ftw.solr.patches.mangler import mangleQuery
from ftw.solr.patches.mangler import mangle_searchable_text_query
from ftw.solr.patches.mangler import searchterms_from_value
from ftw.solr.patches.mangler import trailing_wildcards
from unittest import TestCase
from zope.component import provideUtility, getUtility


class TestQueryMangler(TestCase):

    def setUp(self):
        provideUtility(SolrConnectionConfig(), ISolrConnectionConfig)
        xml = getData('plone_schema.xml')
        xml = xml[xml.find('<schema'):]
        self.schema = SolrSchema(xml.strip())

    def test_search_pattern_value_is_lowercase(self):
        config = getUtility(ISolrConnectionConfig)
        config.search_pattern = 'Title:{value}'

        query = dict(SearchableText='Pass')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['Title:pass'])},
            query
        )

        query = dict(SearchableText='Pass*')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['Title:pass'])},
            query
        )

        query = dict(SearchableText='Pass port')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['Title:(pass port)'])},
            query
        )

    def test_simple_terms_result_in_value_without_wildcards(self):
        config = getUtility(ISolrConnectionConfig)
        config.search_pattern = '{value}'

        query = dict(SearchableText='foo')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['foo'])},
            query
        )

    def test_simple_search_results_in_value_without_wildcards(self):
        config = getUtility(ISolrConnectionConfig)
        config.search_pattern = '{value}'

        query = dict(SearchableText='foo bar')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['(foo bar)'])},
            query
        )

    def test_complex_search_goes_through_unmodified(self):
        config = getUtility(ISolrConnectionConfig)
        config.search_pattern = '{irrelevant}'

        query = dict(SearchableText='foo AND bar')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': 'foo AND bar'},
            query
        )

    def test_simple_terms_are_handled_the_same_as_simple_searches(self):
        config = getUtility(ISolrConnectionConfig)
        config.search_pattern = '{value} OR {value_lwc} OR {value_twc}'

        # Simple term
        query = dict(SearchableText='foo')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(['foo OR (*foo) OR (foo*)'])},
            query
        )

        # Simple search
        query = dict(SearchableText='foo bar')
        mangleQuery(query, config, self.schema)
        self.assertEquals(
            {'SearchableText': set(
                ['(foo bar) OR (*foo *bar) OR (foo* bar*)'])},
            query
        )


class TestMangleSearchableTextQuery(TestCase):

    def test_simple_terms_value_contains_no_wildcards(self):
        orig_query = 'foo'
        pattern = '{value}'
        mangled_query = mangle_searchable_text_query(orig_query, pattern)
        self.assertEquals(
            'foo',
            mangled_query)

    def test_simple_search_results_in_simple_pattern_substitution(self):
        orig_query = 'foo bar'
        pattern = '{value}'
        mangled_query = mangle_searchable_text_query(orig_query, pattern)
        self.assertEquals(
            '(foo bar)',
            mangled_query)

    def test_simple_search_substitutes_lwc_and_twc_values(self):
        orig_query = 'foo bar'
        pattern = '{value_lwc} OR {value_twc}'
        mangled_query = mangle_searchable_text_query(orig_query, pattern)
        self.assertEquals(
            '(*foo *bar) OR (foo* bar*)',
            mangled_query)

    def test_substituted_value_never_contains_any_wildcards(self):
        # Simple term
        orig_query = '*f*oo*'
        pattern = '{value}'
        mangled_query = mangle_searchable_text_query(orig_query, pattern)
        self.assertEquals('foo', mangled_query)

        # Simple search
        orig_query = 'foo* *bar'
        pattern = '{value}'
        mangled_query = mangle_searchable_text_query(orig_query, pattern)
        self.assertEquals('(foo bar)', mangled_query)


class TestWildcardFunctions(TestCase):

    def test_searchterms_from_value_strips_wildcards(self):
        value = '*foo bar* b*az'
        terms = searchterms_from_value(value)
        self.assertEquals(['foo', 'bar', 'baz'], terms)

    # XXX: Output changed due changes in quote method from collective.solr.
    # The queryparser now recognizes boolean operators, therefore it's no
    # longer escaped.
    def test_searchterms_from_value_quotes_terms(self):
        value = 'foo&&bar'
        terms = searchterms_from_value(value)
        self.assertEquals(['foo&&bar'], terms)

    def test_searchterms_from_value_strips_parentheses(self):
        value = '(foo bar)'
        terms = searchterms_from_value(value)
        self.assertEquals(['foo', 'bar'], terms)

        value = ')foo bar('
        terms = searchterms_from_value(value)
        self.assertEquals(['foo', 'bar'], terms)

    def test_leading_wildcards_prepends_wildcards_to_all_terms(self):
        terms = 'foo bar qux'
        result = leading_wildcards(terms)
        self.assertEquals('(*foo *bar *qux)', result)

    def test_leading_wildcards_drops_existing_wildcards(self):
        terms = '*foo bar* qu*x'
        result = leading_wildcards(terms)
        self.assertEquals('(*foo *bar *qux)', result)

    def test_leading_wildcards_adds_parentheses(self):
        terms = 'foo bar'
        result = leading_wildcards(terms)
        self.assertTrue(result.startswith('('))
        self.assertTrue(result.endswith(')'))

    def test_trailing_wildcards_prepends_wildcards_to_all_terms(self):
        terms = 'foo bar qux'
        result = trailing_wildcards(terms)
        self.assertEquals('(foo* bar* qux*)', result)

    def test_trailing_wildcards_drops_existing_wildcards(self):
        terms = '*foo bar* qu*x'
        result = trailing_wildcards(terms)
        self.assertEquals('(foo* bar* qux*)', result)

    def test_trailing_wildcards_adds_parentheses(self):
        terms = 'foo bar'
        result = trailing_wildcards(terms)
        self.assertTrue(result.startswith('('))
        self.assertTrue(result.endswith(')'))


class TestQueryParameters(TestCase):

    def test_keep_valid_facet_fields(self):
        schema = SolrSchema()
        schema['foo'] = SolrField(indexed=True)
        schema['bar'] = SolrField(indexed=True)
        params = cleanupQueryParameters({'facet.field': 'foo'}, schema)
        self.assertEquals({'facet': 'true', 'facet.field': ['foo']}, params)
        params = cleanupQueryParameters({'facet.field': ['foo', 'bar']},
                                        schema)
        self.assertEquals({'facet': 'true', 'facet.field': ['foo', 'bar']},
                          params)

    def test_remove_invalid_facet_fields(self):
        schema = SolrSchema()
        params = cleanupQueryParameters({'facet.field': 'foo'}, schema)
        self.assertEquals({}, params)
        params = cleanupQueryParameters({'facet.field': ['foo', 'bar']},
                                        schema)
        self.assertEquals({}, params)

        schema['foo'] = SolrField(indexed=False)
        params = cleanupQueryParameters({'facet.field': 'foo'}, schema)
        self.assertEquals({}, params)
        params = cleanupQueryParameters({'facet.field': ['foo', 'bar']},
                                        schema)
        self.assertEquals({}, params)

    def test_remove_invalid_and_keep_valid_facet_fields(self):
        schema = SolrSchema()
        schema['bar'] = SolrField(indexed=True)
        params = cleanupQueryParameters({'facet.field': ['foo', 'bar']},
                                        schema)
        self.assertEquals({'facet': 'true', 'facet.field': ['bar']}, params)

    def test_insert_default_qt_parameter(self):
        self.assertEquals({'qt': 'select'}, subtractQueryParameters({}))

    def test_keep_existing_qt_parameter(self):
        self.assertEquals({'qt': 'search'}, subtractQueryParameters(
            {'qt': 'search'}))



# Modified implementation from c.solr.utils:
# - simple terms may contain digits
# - simple terms may contain dots

from re import compile, UNICODE

simpleTerm = compile(r'^[\w\d\.]+$', UNICODE)
def isSimpleTerm(term):
    if isinstance(term, str):
        term = unicode(term, 'utf-8', 'ignore')
    return bool(simpleTerm.match(term.strip()))


operators = compile(r'(.*)\s+(AND|OR|NOT)\s+', UNICODE)
simpleCharacters = compile(r'^[\w\d\?\*\s\.]+$', UNICODE)
def isSimpleSearch(term):
    term = term.strip()
    if isinstance(term, str):
        term = unicode(term, 'utf-8', 'ignore')
    if not term:
        return False

    num_quotes = term.count('"')
    if num_quotes > 0:
        # We consider all queries containing quotes non-simple, whether
        # quotes are balanced or not
        return False

    if bool(operators.match(term)):
        return False
    if bool(simpleCharacters.match(term)):
        return True
    term = term.strip()
    if not term:
        return True
    return False



from ftw.builder import builder_registry
from ftw.builder.dexterity import DexterityBuilder


class DexterityFolderBuilder(DexterityBuilder):
    portal_type = 'DexterityFolder'


builder_registry.register('dexterity folder', DexterityFolderBuilder)



from archetypes.schemaextender.interfaces import ISchemaExtender
from collective.solr import extender
from ftw.solr import _
from Products.Archetypes.atapi import BooleanWidget
from Products.Archetypes.atapi import TextAreaWidget
from zope.interface import implements


class SearchExtender(object):
    """Adapter that adds search metadata."""

    implements(ISchemaExtender)

    _fields = [
        extender.ExtensionBooleanField(
            'showinsearch',
            languageIndependent=True,
            schemata='settings',
            default=True,
            write_permission='ftw.solr: Edit search settings',
            widget=BooleanWidget(
                label=_(u'label_showinsearch',
                        default=u'Show in search'),
                visible={"edit": "visible", "view": "invisible"},
                description="",
                )),

        extender.ExtentionTextField(
            'searchwords',
            searchable=False,
            schemata='settings',
            languageIndependent=False,
            write_permission='ftw.solr: Edit search settings',
            allowable_content_types = 'text/plain',
            widget=TextAreaWidget(
                label=_(u'label_searchwords', default=u'Search words'),
                description=_(
                    u'help_searchwords',
                    default=u'Specify words for which this item will show up '
                    'as the first search result. Multiple words can be '
                    'specified on new lines.'),
                visible={"edit": "visible", "view": "invisible"},
                )),
        ]

    def __init__(self, context):
        self.context = context

    def getFields(self):
        return self._fields



from Acquisition import aq_inner
from ftw.solr import _
from ftw.solr.interfaces import ILiveSearchSettings
from itertools import groupby
from plone import api
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.browser.navtree import getNavigationRoot
from Products.CMFPlone.utils import safe_unicode
from Products.Five.browser import BrowserView
from Products.PythonScripts.standard import html_quote
from Products.PythonScripts.standard import url_quote_plus
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.i18n import translate
import json


def quotestring(s):
    return '"%s"' % s


def quote_bad_chars(s):
    bad_chars = ["(", ")"]
    for char in bad_chars:
        s = s.replace(char, quotestring(char))
    return s


class FtwSolrLiveSearchReplyView(BrowserView):

    def __call__(self):
        context = aq_inner(self.context)
        q = self.request.form.get('term', None)
        self.limit = int(self.request.form.get('limit', 10))
        path = self.request.form.get('path', None)

        plone_utils = getToolByName(context, 'plone_utils')
        self.pretty_title_or_id = plone_utils.pretty_title_or_id
        self.normalizeString = plone_utils.normalizeString
        plone_view = getMultiAdapter((context, self.request), name='plone')
        self.get_icon = plone_view.getIcon

        pprops = getToolByName(context, 'portal_properties')
        sprops = getattr(pprops, 'site_properties', None)
        self.useViewAction = []
        if sprops is not None:
            self.useViewAction = sprops.getProperty('typesUseViewActionInListings',
                                                    [])

        registry = getUtility(IRegistry)
        self.settings = registry.forInterface(ILiveSearchSettings)

        # XXX really if it contains + * ? or -
        # it will not be right since the catalog ignores all non-word
        # characters equally like
        # so we don't even attept to make that right.
        # But we strip these and these so that the catalog does
        # not interpret them as metachars
        # See http://dev.plone.org/plone/ticket/9422 for an explanation of
        # '\u3000'
        multispace = u'\u3000'.encode('utf-8')
        for char in ('?', '-', '+', '*', multispace):
            q = q.replace(char, ' ')
        r = quote_bad_chars(q) + '*'
        self.searchterms = url_quote_plus(r)

        site_encoding = plone_utils.getSiteEncoding()
        if path is None:
            path = getNavigationRoot(context)
        catalog = getToolByName(context, 'portal_catalog')
        friendly_types = plone_utils.getUserFriendlyTypes()

        self.facet_params = context.restrictedTraverse(
            '@@search-facets/facet_parameters')()

        self.searchterm_query = '?searchterm=%s' % url_quote_plus(q)

        if self.settings.grouping:
            results = catalog(SearchableText=r, portal_type=friendly_types,
                              request_handler='livesearch', path=path,
                              sort_limit=self.settings.group_search_limit)[:self.settings.group_search_limit]

        else:
            results = catalog(SearchableText=r, portal_type=friendly_types,
                              request_handler='livesearch', path=path,
                              sort_limit=self.limit)[:self.limit]

        self.request.response.setHeader(
            'Content-Type', 'application/json;charset=%s' % site_encoding)

        return self.get_results_as_json(results)

    def get_results_as_json(self, results):
        if self.settings.grouping:
            grouped_results = []

            group_func = lambda result: result['portal_type']

            def sort_func(item):
                group_by = self.settings.group_by
                try:
                    return group_by.index(item['portal_type'])
                except ValueError:
                    return len(group_by) + 1

            results = sorted(results, key=sort_func)

            group_limit = self.settings.group_limit

            for portal_type, group in groupby(results, group_func):

                if group_limit:
                    group = list(group)[:self.settings.group_limit]
                else:
                    group = list(group)

                group[0]['first_of_group'] = True
                grouped_results.extend(group)

            return json.dumps(self.generate_items(grouped_results))

        else:
            return json.dumps(self.generate_items(results))

    def generate_items(self, results):
        data = []
        for result in results:
            url = result.getURL()
            if result.portal_type in self.useViewAction:
                url += '/view'
            url = url + self.searchterm_query
            title = self.get_title(result)
            first_of_group = 'first_of_group' in result and result.first_of_group

            # XXX Typestools lookup could be cacheable
            types_tool = api.portal.get_tool('portal_types')

            type_domain = types_tool.get(result.portal_type).i18n_domain

            # XXX Somehow solr does not return the "Type" for Files - The
            # binary adder explicitly ignores this attr.
            if result.portal_type == 'File':
                portal_type = translate(result.portal_type,
                                        domain=type_domain,
                                        context=self.request)
            else:
                portal_type = translate(result.Type,
                                        domain=type_domain,
                                        context=self.request)

            data.append({'icon': self.get_icon(result).html_tag() or '',
                         'url': url,
                         'title': title,
                         'description': self.get_description(result),
                         'cssclass': self.get_css_class(result),
                         'type': portal_type,
                         'firstOfGroup': first_of_group
                         })

        if(len(data)):
            data.append(self.get_show_more_item())
        else:
            data.append(self.get_nothing_found_item())

        search_help_link = self.get_search_help_item()
        if search_help_link:
            data.append(self.get_search_help_item())

        return data

    def get_title(self, brain):
        return safe_unicode(self.pretty_title_or_id(brain))

    def get_description(self, brain):
        display_description = safe_unicode(brain.Description)
        # need to quote it, to avoid injection of html containing javascript
        # and other evil stuff
        return html_quote(display_description)

    def get_css_class(self, brain):
        return 'contenttype-{0}'.format(
            self.normalizeString(brain.portal_type))

    def get_search_help_item(self):
        label_search_help = _('label_search_help', default='Search help')
        search_help_obj = api.portal.get_navigation_root(
            self.context).restrictedTraverse('searchhelp', None)

        if search_help_obj is None:
            return None

        return {
            'url': search_help_obj.absolute_url(),
            'title': translate(label_search_help, context=self.request),
            'first_of_group': False,
            'cssclass': 'no-result'
            }

    def get_show_more_item(self):
        label_show_all = _('label_show_all', default='Show all items')
        params = self.facet_params
        params += '&SearchableText=' + self.searchterms
        path = self.request.form.get('path', None)
        if path:
            params += '&path=' + url_quote_plus(path)
        url = '@@search?{0}'.format(params)

        return {
            'url': url,
            'title': translate(label_show_all, context=self.request),
            'first_of_group': False,
            'cssclass': 'no-result'
        }

    def get_nothing_found_item(self):
        label_nothing_found = _('label_nothing_found',
                                default='No items found')
        params = self.facet_params
        params += '&SearchableText=' + self.searchterms
        path = self.request.form.get('path', None)
        if path:
            params += '&path=' + url_quote_plus(path)
        url = '@@search?{0}'.format(params)

        return {
            'url': url,
            'title': translate(label_nothing_found, context=self.request),
            'first_of_group': False,
            'cssclass': 'no-result'
        }



from plone import api
from plone.indexer import indexer
from plone.indexer.interfaces import IIndexer
from plone.indexer.wrapper import IndexableObjectWrapper
from Products.Archetypes.interfaces.base import IBaseObject
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.utils import getToolByName
from Products.PluginIndexes.common import safe_callable
from Products.ZCatalog.interfaces import IZCatalog
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.interface import implements
from zope.interface import Interface
import re


@indexer(Interface)
def site_area(obj):
    """ Index the title of the closest navigation root
    """
    return api.portal.get_navigation_root(obj).Title()


@indexer(Interface)
def site_section(obj, **kwargs):
    """An indexer which returns the section title for a content item.
    """
    portal_state = getMultiAdapter((obj, obj.REQUEST),
                                   name=u'plone_portal_state')
    navroot = portal_state.navigation_root()
    contentPath = obj.getPhysicalPath()[len(navroot.getPhysicalPath()):]
    if contentPath:
        section = navroot.unrestrictedTraverse(contentPath[0])
        return section.Title()
    return ''


@indexer(Interface)
def creator_fullname(obj, **kwargs):
    """An indexer which returns the full name of the content creator.
    """
    creator = obj.Creator()
    mtool = getToolByName(obj, 'portal_membership')
    member_info = mtool.getMemberInfo(creator)
    if member_info and member_info['fullname']:
        return member_info['fullname']
    return creator


class SnippetTextIndexer(object):
    """Text for snippets (aka highlighting) in search results.
       Uses the SearchableText but excludes some fields that should not be
       shown in search results.
    """
    implements(IIndexer)
    adapts(Interface, IZCatalog)

    def __init__(self, context, catalog):
        self.context = context
        self.catalog = catalog

    def __call__(self):
        wrapped = IndexableObjectWrapper(self.context, self.catalog)
        text = getattr(wrapped, 'SearchableText')
        if safe_callable(text):
            text = text()

        # Archetypes object: remove id and title
        if IBaseObject.providedBy(self.context):
            for fieldname in ['id', 'title']:
                field = self.context.Schema().getField(fieldname)
                if field is None:
                    continue

                method = field.getIndexAccessor(self.context)
                value = method()
                text = text.replace(value, '', 1)
        # other content (e.g. dexterity): remove title
        elif IContentish.providedBy(self.context):
            text = text.replace(self.context.Title(), '', 1)

        # Strip html tags
        text = re.sub('<[^<]+?>', '', text)

        return text
